/* keyconv.c - Converts Linux keymaps to Retro format

Copyright (C) 1999 Thomas Novelli
	Retro is distributed under the terms of the GNU LGPL, version 2 or
	greater. (For the text of the license, see the file COPYING.LIB)

Usage note:
	You'll see some "Unknown keysysm" messages when you run keyconv.
	You can probably ignore them, but check them to be sure.  If some
	keys don't work under Retro, this is probably the problem.
	If you need to add a keysym, add these two things:
		1. A keysym handler function
		2. An entry in tokens[]
	Just search for "// Keysym handlers" and you'll see where to add
	them.  I hope it's obvious enough for you.


Note:
	I could have used lex/yacc to parse the keymaps, but I didn't.  I
	don't like lex/yacc, I'd like something better, and this here is an
	experiment.  It does the job just fine, by the way.  So don't bug me
	about it! :)

To do:
	Finish keysym handlers:
		Caps_Lock, Num_Lock, Scroll_Lock... handle these as functions?
		Caps_On, Uncaps_Shift (for typewriter-like CapsLock)
		Pause, F-keys, Arrows, Keypad
		Console switching?, Scrolling?
		Certain control- and meta- combinations (like ^^ and ^_)
	Handle 'compose' and 'dead keys'
	Handle 'include'

Not important:
	Handle strings (lots of work, for little benefit)
*/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>


void parse(FILE *f);
void parse_word(char *s);
void keymaps();
void keymaplist(char *s);
void keycode();
void scancode(char *s);
void equals();
void symbol(int type, int data);

void string();
void compose();
void to();

void dump();
void output(FILE *f);


struct TABLE {
  int t;	// type
  int d;	// data
};

struct TABLE table[128][256];	// [row][column]
//struct TABLE **table;		// dynamic-allocated (had problems with this)

int cols[256], ncols = 0;

int col_mask;		// OR-mask of current modifiers
int plain;		// 'plain' modifier active?

int cur_scancode;	// Scancode & column to be filled in by the next
int cur_column;		// keycode definition (index into table)

void m_plain()		{ plain = 1; }
void m_shift()		{ col_mask += 1; }
void m_altgr()		{ col_mask += 2; }
void m_control()	{ col_mask += 4; }
void m_alt()		{ col_mask += 8; }
void m_shiftL()		{ col_mask += 16; }
void m_shiftR()		{ col_mask += 32; }
void m_ctrlL()		{ col_mask += 64; }
void m_ctrlR()		{ col_mask += 128; }



// Action types
#define A_FUNCTION	0	// Function call
#define A_ASCII		1
#define A_STRING	2
#define A_MODIFIER	3



// Keysym handlers
/* Templates (copy and fill in above the *):
-- ASCII character:
void s_()		{ symbol(A_ASCII, ''); }
       *				  * ASCII code

-- Function call:
void s_()		{ static char s[] = ""; symbol(A_FUNCTION, (int)s); }
       *				     * Name of function in Retro
*/ 
void s_escape()		{ symbol(A_ASCII, 27); }
void s_boot()		{ static char s[] = "reboot"; symbol(A_FUNCTION, (int)s); }
void s_0()		{ symbol(A_ASCII, '0'); }	// top row
void s_1()		{ symbol(A_ASCII, '1'); }
void s_2()		{ symbol(A_ASCII, '2'); }
void s_3()		{ symbol(A_ASCII, '3'); }
void s_4()		{ symbol(A_ASCII, '4'); }
void s_5()		{ symbol(A_ASCII, '5'); }
void s_6()		{ symbol(A_ASCII, '6'); }
void s_7()		{ symbol(A_ASCII, '7'); }
void s_8()		{ symbol(A_ASCII, '8'); }
void s_9()		{ symbol(A_ASCII, '9'); }
void s_minus()		{ symbol(A_ASCII, '-'); }
void s_equal()		{ symbol(A_ASCII, '='); }
void s_exclam()		{ symbol(A_ASCII, '!'); }	// top row, shifted
void s_at()		{ symbol(A_ASCII, '@'); }
void s_numbersign()	{ symbol(A_ASCII, '#'); }
void s_dollar()		{ symbol(A_ASCII, '$'); }
void s_percent()	{ symbol(A_ASCII, '%'); }
void s_circum()		{ symbol(A_ASCII, '^'); }
void s_ampersand()	{ symbol(A_ASCII, '&'); }
void s_asterisk()	{ symbol(A_ASCII, '*'); }
void s_parenleft()	{ symbol(A_ASCII, '('); }
void s_parenright()	{ symbol(A_ASCII, ')'); }
void s_underscore()	{ symbol(A_ASCII, '_'); }
void s_plus()		{ symbol(A_ASCII, '+'); }
void s_nul()		{ symbol(A_ASCII, 0); }		// 'special keys'
void s_tab()		{ symbol(A_ASCII, 9); }
void s_delete()		{ symbol(A_ASCII, 8); }
void s_return()		{ symbol(A_ASCII, 10); }
void s_space()		{ symbol(A_ASCII, 32); }
//void s_pause()		{ static char s[] = "pause"; symbol(A_FUNCTION, (int)s); }
void s_bracketleft()	{ symbol(A_ASCII, '['); }	// punctuation
void s_bracketright()	{ symbol(A_ASCII, ']'); }
void s_braceleft()	{ symbol(A_ASCII, '{'); }
void s_braceright()	{ symbol(A_ASCII, '}'); }
void s_period()		{ symbol(A_ASCII, '.'); }
void s_comma()		{ symbol(A_ASCII, ','); }
void s_less()		{ symbol(A_ASCII, '<'); }
void s_greater()	{ symbol(A_ASCII, '>'); }
void s_slash()		{ symbol(A_ASCII, '/'); }
void s_question()	{ symbol(A_ASCII, '?'); }
void s_semicolon()	{ symbol(A_ASCII, ';'); }
void s_colon()		{ symbol(A_ASCII, ':'); }
void s_apostrophe()	{ symbol(A_ASCII, '\''); }
void s_quotedbl()	{ symbol(A_ASCII, '\"'); }
void s_backslash()	{ symbol(A_ASCII, '\\'); }
void s_bar()		{ symbol(A_ASCII, '|'); }
void s_tilde()		{ symbol(A_ASCII, '~'); }
void s_grave()		{ symbol(A_ASCII, '`'); }

void s_shift()		{ symbol(A_MODIFIER, 1); }	// modifiers
void s_altgr()		{ symbol(A_MODIFIER, 2); }
void s_control()	{ symbol(A_MODIFIER, 4); }
void s_alt()		{ symbol(A_MODIFIER, 8); }
void s_shiftL()		{ symbol(A_MODIFIER, 16); }
void s_shiftR()		{ symbol(A_MODIFIER, 32); }
void s_ctrlL()		{ symbol(A_MODIFIER, 64); }
void s_ctrlR()		{ symbol(A_MODIFIER, 128); }

void s_up()		{ symbol(A_ASCII, 0x0100); }	// arrows, etc.
void s_left()		{ symbol(A_ASCII, 0x0200); }
void s_right()		{ symbol(A_ASCII, 0x0300); }
void s_down()		{ symbol(A_ASCII, 0x0400); }
void s_ins()		{ symbol(A_ASCII, 0x0500); }
void s_del()		{ symbol(A_ASCII, 0x0600); }
void s_home()		{ symbol(A_ASCII, 0x0700); }
void s_end()		{ symbol(A_ASCII, 0x0800); }
void s_pgup()		{ symbol(A_ASCII, 0x0900); }
void s_pgdn()		{ symbol(A_ASCII, 0x0a00); }
void s_f1()		{ symbol(A_ASCII, 0x1000); }	// function keys
void s_f2()		{ symbol(A_ASCII, 0x1100); }
void s_f3()		{ symbol(A_ASCII, 0x1200); }
void s_f4()		{ symbol(A_ASCII, 0x1300); }
void s_f5()		{ symbol(A_ASCII, 0x1400); }
void s_f6()		{ symbol(A_ASCII, 0x1500); }
void s_f7()		{ symbol(A_ASCII, 0x1600); }
void s_f8()		{ symbol(A_ASCII, 0x1700); }
void s_f9()		{ symbol(A_ASCII, 0x1800); }
void s_f10()		{ symbol(A_ASCII, 0x1900); }
void s_f11()		{ symbol(A_ASCII, 0x1a00); }
void s_f12()		{ symbol(A_ASCII, 0x1b00); }

void s_macro()		{ symbol(A_ASCII, 0xf000); }	// other
void s_voidsym()	{ symbol(A_FUNCTION, 0); }


// Token types
#define T_KEYMAPS	1
#define T_KEYMAPLIST	11
#define T_KEYCODE	2
#define T_MOD		3
#define T_SYM		4
#define T_EQUALS	5
#define T_SCANCODE	6
#define T_STRING	13
#define T_STRING_NAME	7
#define T_STRING_DATA	12
#define	T_COMPOSE	8
#define T_CHAR		9
#define	T_TO		10

struct TOKEN {
  char *s;	// string
  void (*f)();	// handler function
  int t;	// type (for checking validity)
};

struct TOKEN tokens[] = {
  { "keymaps",	keymaps,	T_KEYMAPS },
  { "keycode",	keycode,	T_KEYCODE },
  { "=",	equals,		T_EQUALS },

  { "plain",	m_plain,	T_MOD },
  { "shift",	m_shift,	T_MOD },
  { "altgr",	m_altgr,	T_MOD },
  { "control",	m_control,	T_MOD },
  { "alt",	m_alt,		T_MOD },
  { "shiftl",	m_shiftL,	T_MOD },
  { "shiftr",	m_shiftR,	T_MOD },
  { "ctrll",	m_ctrlL,	T_MOD },
  { "ctrlr",	m_ctrlR,	T_MOD },

  { "string",	string,		T_STRING },
  { "compose",	compose,	T_COMPOSE },
  { "to",	to,		T_TO },

  // Keysym handlers
/* Template entry:
  { "",	s_,		T_SYM },
*/
  { "Escape",	s_escape,	T_SYM },
  { "Boot",	s_boot,		T_SYM },
  { "zero",	s_0,		T_SYM },	// Digits
  { "one",	s_1,		T_SYM },
  { "two",	s_2,		T_SYM },
  { "three",	s_3,		T_SYM },
  { "four",	s_4,		T_SYM },
  { "five",	s_5,		T_SYM },
  { "six",	s_6,		T_SYM },
  { "seven",	s_7,		T_SYM },
  { "eight",	s_8,		T_SYM },
  { "nine",	s_9,		T_SYM },
  { "minus",	s_minus,	T_SYM },
  { "plus",	s_plus,		T_SYM },
  { "exclam",		s_exclam,	T_SYM },	// Shifted digits
  { "at",		s_at,		T_SYM },
  { "numbersign",	s_numbersign,	T_SYM },
  { "dollar",		s_dollar,	T_SYM },
  { "percent",		s_dollar,	T_SYM },
  { "asciicircum",	s_circum,	T_SYM },
  { "ampersand",	s_ampersand,	T_SYM },
  { "asterisk",		s_asterisk,	T_SYM },
  { "parenleft",	s_parenleft,	T_SYM },
  { "parenright",	s_parenright,	T_SYM },
  { "underscore",	s_underscore,	T_SYM },
  { "equal",		s_equal,	T_SYM },
  { "nul",	s_nul,		T_SYM },		// "special keys"
  { "Tab",	s_tab,		T_SYM },
  { "Return",	s_return,	T_SYM },
  { "Delete",	s_delete,	T_SYM },
  { "space",	s_space,	T_SYM },
//  { "Pause",	s_pause,	T_SYM },
  { "bracketleft",	s_bracketleft,	T_SYM },	// punctuation
  { "bracketright",	s_bracketright,	T_SYM },
  { "braceleft",	s_braceleft,	T_SYM },
  { "braceright",	s_braceright,	T_SYM },
  { "comma",		s_comma,	T_SYM },
  { "less",		s_less,		T_SYM },
  { "period",		s_period,	T_SYM },
  { "greater",		s_greater,	T_SYM },
  { "slash",		s_slash,	T_SYM },
  { "question",		s_question,	T_SYM },
  { "backslash",	s_backslash,	T_SYM },
  { "bar",		s_bar,		T_SYM },
  { "apostrophe",	s_apostrophe,	T_SYM },
  { "quotedbl",		s_quotedbl,	T_SYM },
  { "colon",		s_colon,	T_SYM },
  { "semicolon",	s_semicolon,	T_SYM },
  { "asciitilde",	s_tilde,	T_SYM },
  { "grave",		s_grave,	T_SYM },
  { "Shift",		s_shift,	T_SYM },	// modifiers
  { "AltGr",		s_altgr,	T_SYM },
  { "Control",		s_control,	T_SYM },
  { "Alt",		s_alt,		T_SYM },
  { "ShiftL",		s_shiftL,	T_SYM },
  { "ShiftR",		s_shiftR,	T_SYM },
  { "CtrlL",		s_ctrlL,	T_SYM },
  { "CtrlR",		s_ctrlR,	T_SYM },

  { "Up",	s_up,		T_SYM },		// arrows & other editing keys
  { "Left",	s_left,		T_SYM },
  { "Right",	s_right,	T_SYM },
  { "Down",	s_down,		T_SYM },
  { "Insert",	s_ins,		T_SYM },
  { "Remove",	s_del,		T_SYM },
  { "Find",	s_home,		T_SYM },
  { "Select",	s_end,		T_SYM },
  { "Prior",	s_pgup,		T_SYM },
  { "Next",	s_pgdn,		T_SYM },

  { "KP_Enter",		s_return,	T_SYM },	// keypad keys
  { "KP_Divide",	s_slash,	T_SYM },	// (these keys should really be
  { "KP_Multiply",	s_asterisk,	T_SYM },	// handled differently from normal keys)
  { "KP_Period",	s_period,	T_SYM },
  { "KP_Add",		s_plus,		T_SYM },
  { "KP_Subtract",	s_minus,	T_SYM },
  { "KP_1",	s_1,		T_SYM },
  { "KP_2",	s_2,		T_SYM },
  { "KP_3",	s_3,		T_SYM },
  { "KP_4",	s_4,		T_SYM },
  { "KP_5",	s_5,		T_SYM },
  { "KP_6",	s_6,		T_SYM },
  { "KP_7",	s_7,		T_SYM },
  { "KP_8",	s_8,		T_SYM },
  { "KP_9",	s_9,		T_SYM },
  { "KP_0",	s_0,		T_SYM },
  { "F1",	s_f1,		T_SYM },		// function keys
  { "F2",	s_f2,		T_SYM },
  { "F3",	s_f3,		T_SYM },
  { "F4",	s_f4,		T_SYM },
  { "F5",	s_f5,		T_SYM },
  { "F6",	s_f6,		T_SYM },
  { "F7",	s_f7,		T_SYM },
  { "F8",	s_f8,		T_SYM },
  { "F9",	s_f9,		T_SYM },
  { "F10",	s_f10,		T_SYM },
  { "F11",	s_f11,		T_SYM },
  { "F12",	s_f12,		T_SYM },
//  { "Macro",	s_macro,		T_SYM },	// other
  { "VoidSymbol",	s_voidsym,	T_SYM },
  { 0 }
};


// State machine
#define ST_NORMAL	1	// Valid: T_KEYMAPS, T_KEYCODE, T_STRING, T_COMPOSE,
				//	  eol
#define ST_EQUALS	2	// Valid: T_EQUALS
#define ST_SCANCODE	3	// Valid: T_SCANCODE
#define ST_SYMBOL	4	// Valid: T_SYMBOL, eol
#define ST_COMPOSE_X	6	// Valid: T_CHAR
#define ST_COMPOSE_Y	7	// Valid: T_CHAR
#define ST_TO		8	// Valid: T_TO
#define ST_COMPOSE_Z 	9	// Valid: T_CHAR
#define ST_STRING_NAME	10	// Valid: T_STRING_NAME
#define ST_STRING_DATA	11	// Valid: T_STRING_DATA
#define ST_KEYMAPLIST	12	// Valid: T_KEYMAPLIST

int state;
int old_state;	// currently only set by equals()



// Parameters: <input filename> <output filename>
int main (int argc, char **argv)
{
  FILE *fi, *fo;

  if (argc != 3) {
    printf("Usage: %s <Linux keymap file> <output file>\n", argv[0]);
    exit(1);
  }

  if (!(fi = fopen(argv[1], "rt"))) {
    printf("Error: Unable to open Linux keymap file: %s\n", argv[1]);
    exit(2);
  }

  if (!(fo = fopen(argv[2], "wb"))) {
    printf("Error: Unable to write output file: %s\n", argv[2]);
    exit(3);
  }

  parse(fi);	// Convert text ==> internal format
  dump();
  output(fo);	// Write out in Retro format (binary)

  fclose(fi);
  fclose(fo);

  return 0;
}


/**************************************************************************/


void parse(FILE *f)
{
  char s[256];
  char *word, *p;
  int i;

  state = ST_NORMAL;

  do {
    fgets(s, 256, f);
//    printf("%s", s);
    
    // Beginning of line:
    state = ST_NORMAL;
    cur_column = col_mask = 0;
    plain = 0;
    
    p = s;
    do {
      while (isspace(*p) && *p) { p++; }	// Skip leading spaces
      word = p;
      while (!isspace(*p) && *p) { p++; }
      if (*p) {		// If (not at end of string) ...
        *p = 0;
        p++;
        parse_word(word);
      }
    } while (*p);
    
    // End of line:
    if (state == ST_SYMBOL) {		// "keycode nnn = " needs special EOL handling
      if (cur_column == 1 && !plain) {
        // Copy column 0 to all columns
        for (i=1; i<ncols; i++)
          table[cur_scancode][i] = table[cur_scancode][0];
        // For letters, do a few things automatically
        if (table[cur_scancode][0].t == A_ASCII)
          if (islower(table[cur_scancode][0].d)) {
            table[cur_scancode][1].d = toupper(table[cur_scancode][0].d);	// Shift
            table[cur_scancode][3].d = table[cur_scancode][0].d - 'a' + 1;	// Control
          }
      }
    }
  } while (!feof(f));
}


void parse_word(char *s)
{
  int i = 0;
  int T;	// type
  int ok;

//  printf("%d: %s\n", strlen(s), s);
  while (tokens[i].s) {
    if (!strcmp(s, tokens[i].s))
      break;
    i++;
  }
  if (tokens[i].s) {	// Matched...
//    printf(" *** Matched: %d %p %s\n", tokens[i].t, tokens[i].f, tokens[i].s);

    // Todo: Check to see if this token is allowed in the current state
    T = tokens[i].t;
    ok=0;
    switch (state) {
    case ST_NORMAL:
      if (T==T_KEYMAPS || T==T_MOD || T==T_KEYCODE || T==T_STRING || T==T_COMPOSE) ok=1;
      break;
    case ST_EQUALS:
      if (T==T_EQUALS) ok=1; 
      break;
    case ST_TO:
      if (T==T_TO) ok=1;
      break;
    case ST_SYMBOL:
      if (T==T_SYM) ok=1;
      break;
    }
//    printf("state=%d  ok=%d\n", state, ok);
    if (ok) {
      if (tokens[i].f)
        (tokens[i].f)();
    }
    else {
      printf("Error: \"%s\" not allowed here!\n", s);
      // Todo: say what was expected... (string for current state)
    }
  }
  else {		// Could be a number or something...
    switch (state) {
    case ST_KEYMAPLIST:		keymaplist(s); break;
    case ST_SCANCODE:		scancode(s); break;
    case ST_SYMBOL:
      // Check for special cases
      // ('+a' notation, ASCII letters, etc...)
      if (strlen(s) == 1) {
//        printf("Single char: %c\n", *s);
        symbol(A_ASCII, *s);
      }
      else {
        // If a symbol is unrecognized, consider it a VoidSymbol (i.e., unmap the key)
        printf("Unknown keysym: %s\n", s);
        symbol(A_FUNCTION, 0);
      }
      break;
    case ST_STRING_NAME:	// ignore
      old_state = state;
      state = ST_EQUALS;
      break;
    case ST_STRING_DATA:	// ignore
      break;
    case ST_COMPOSE_X:		// ignore
      state = ST_COMPOSE_Y;
      break;
    case ST_COMPOSE_Y:		// ignore
      state = ST_TO;
      break;
    case ST_COMPOSE_Z:		// ignore
      break;
    default:
      printf(" *** Not matched: %s\n", s);
    }
  }
}


/************************************************************************/


void keymaps()
{
  state = ST_KEYMAPLIST;
}


void keymaplist(char *s)
{
  char *word, *p;
  int start, end, i;
  
  p = s;
  do {
    word = p;
    while (isdigit(*p)) p++;
    switch (*p) {
    case ',':
      *p = 0; p++;
      // drop thru...
    case 0:
      cols[ncols] = atoi(word); ncols++;
      break;
    case '-':
      *p = 0; p++;
      start = atoi(word);
      word = p;
      while (isdigit(*p)) p++;
      if (*p) { *p = 0; p++; }
      end = atoi(word);
      for (i=start; i<=end; i++) {
        cols[ncols] = i; ncols++;
      }
      break;
    }
  } while (*p);
  
  printf("Columns actually used:");
  for (i=0; i<ncols; i++)
    printf(" %d", cols[i]);
  printf("\n");
  
  // Allocate storage for the table: ncols * 128 scancodes * 8 bytes per entry
//  table = (void *) malloc(ncols * 128 * 8);
  
  state = ST_NORMAL;
}


void keycode()
{
  state = ST_SCANCODE;
}


void scancode(char *s)
{
  cur_scancode = strtol(s, (char **)0, 0);
  if (cur_scancode == LONG_MIN || cur_scancode == LONG_MAX) {
    printf("Error: scancode expected; got \"%s\" instead!\n", s);
    cur_scancode = 0;
  }
  old_state = state;
  state = ST_EQUALS;
}


void equals()
{
  switch (old_state) {
  case ST_SCANCODE:	state = ST_SYMBOL; break;
  case ST_STRING_NAME:	state = ST_STRING_DATA; break;
  }
}


void symbol(int type, int data)
{
  int j;

  // Map symbol w/ specified action type and data
  if (col_mask) {	// Decode col_mask to cur_column
    for (j=0; j<ncols; j++) {
      if (col_mask == cols[j])
        break;
    }
    if (j == ncols) {
      printf("Error: modifier combination wasn't specified in the 'keymaps' line!\n");
    }
    else
      cur_column = j;
  }

//  printf(" *** symbol(): plain=%d  col_mask=%d  [r,c]=[%d,%d] t=%d d=%u\n",
//      		plain, col_mask, cur_scancode, cur_column, type, data);

//  if (type == A_FUNCTION && data)
//    printf("Function: \"%s\" in Retro\n", data);

  table[cur_scancode][cur_column].t = type;
  table[cur_scancode][cur_column].d = data;
  cur_column++;
  // No change in state
}




/***************************** String & Compose *************************/

void string()
{
  state = ST_STRING_NAME;
}


void compose()
{
  state = ST_COMPOSE_X;
}


void to()
{
  state = ST_COMPOSE_Z;
}


/***************************** Output *********************************/

void dump()
{
  int r, c;	// row, col
  int t, d;	// type, data
  
  printf("\nKeymap dump:\n");
  for (c=0; c<ncols; c++) {
    printf("\t%d", cols[c]);
  }
  printf("\n");
  
  for (r=1; r<128; r++) {
    printf("%3d -", r);
    for (c=0; c<ncols; c++) {
      printf("\t");
      t = table[r][c].t;
      d = table[r][c].d;
      if (t == A_ASCII) {
        if (d < 256) {
          if (isprint(d))
            printf("'%c'", d);
          else
            printf("%d", d);
        }
        else {
          printf("0x%04x", d);
        }
      }
      else if (t == A_MODIFIER) {
        printf("+%d", d);
      }
      else if (t == A_FUNCTION || t == A_STRING) {
        if (d)
          printf("%s", (char*)d);
      }
    }
    printf("\n");
  }
}


void output(FILE *f)
{
  int r, c;	// row, col
  int t, d;	// type, data
  int offset = 0;
  int tmp;

  fseek(f, 8, SEEK_CUR);

  // Output key tables
  for (c=0; c<ncols; c++) {
    for (r=0; r<128; r++) {
      t = table[r][c].t;
      fwrite(&t, 4, 1, f);
      d = table[r][c].d;
      if (t == A_ASCII || t ==  A_MODIFIER) {
        fwrite(&d, 4, 1, f);
      }
      else if (t == A_FUNCTION || t == A_STRING) {
        if (d) {
          tmp = offset;
          offset += strlen((char*)d) + 1;
          d = tmp;
        }
        fwrite(&d, 4, 1, f);
      }
    }
  }
  
  // Output string table
  for (c=0; c<ncols; c++) {
    for (r=0; r<128; r++) {
      t = table[r][c].t;
      d = table[r][c].d;
      if (t == A_FUNCTION || t == A_STRING) {
        if (d) {
          fputs((char*)d, f);
          fputc(0, f);
        }
      }
    }
  }

  // Output cols[0..ncols-1]
  for (c=0; c<ncols; c++)
    fputc(cols[c], f);
    
  // Output header
  rewind(f);
  fwrite(&ncols, 4, 1, f);
  fwrite(&offset, 4, 1, f); 
}
